from datetime import datetime
from typing import Callable, List, Optional, Tuple, Union, Dict, Any
from abc import ABC, abstractmethod
import pandas as pd
import pyarrow
import pytz
from pydantic.typing import Literal

from feast import FileSource, OnDemandFeatureView
from feast.data_source import DataSource
from feast.errors import FeastJoinKeysDuringMaterialization
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores.file_source import SavedDatasetFileStorage
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.offline_stores.offline_utils import (
    DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL,
)
from feast.infra.provider import (
    _get_requested_feature_views_to_features_dict,
)
from feast.registry import Registry
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.usage import log_exceptions_and_usage
from feast.importer import import_class
from enum import Enum
import pandas as pd
from datetime import datetime

YUMMY_ALL = "@yummy*"

def select_all(event_timestamp: datetime):
    """
    selects all entities during fetching historical features for specified event timestamp
    """
    return pd.DataFrame.from_dict({YUMMY_ALL: ["*"], DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL: [ event_timestamp ]})


class BackendType(str, Enum):
    dask = "dask"
    ray = "ray"
    spark = "spark"
    polars = "polars"


class BackendConfig(FeastConfigBaseModel):

    backend: Optional[str] = None

    config: Optional[Dict[str, str]] = None
    """ Configuration """


class Backend(ABC):
    """
    Backend implements all operations required to process all offline store steps using
    selected engine
    """
    def __init__(self, backend_config: BackendConfig):
        self._backend_config = backend_config

    @property
    def backend_type(self) -> BackendType:
        raise NotImplementedError("Backend type not defined")

    @property
    def retrival_job_type(self):
        raise NotImplementedError("Retrival job type not defined")

    @abstractmethod
    def first_event_timestamp(
        self,
        entity_df: Union[pd.DataFrame, Any],
        column_name: str
    ) -> datetime:
        """
        Fetch first event timestamp
        """
        ...


    @abstractmethod
    def prepare_entity_df(
        self,
        entity_df: Union[pd.DataFrame, Any],
    ) -> Union[pd.DataFrame, Any]:
        """
        Maps entity_df to type required by backend and finds event timestamp column
        """
        ...

    @abstractmethod
    def normalize_timezone(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        entity_df_event_timestamp_col: str,
    ) -> Union[pd.DataFrame, Any]:
        """
        Normalize timezon of input entity df to UTC
        """
        ...

    @abstractmethod
    def sort_values(
        self,
        entity_df: Union[pd.DataFrame, Any],
        by: Union[str,List[str]],
        ascending: bool = True,
        na_position: Optional[str] = "last",
    ) -> Union[pd.DataFrame, Any]:
        """
        Sorts entity df by selected column
        """
        ...

    @abstractmethod
    def run_field_mapping(
        self,
        table: Union[pd.DataFrame,Any],
        field_mapping: Dict[str, str],
    ):
        ...

    @abstractmethod
    def join(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        df_to_join: Union[pd.DataFrame, Any],
        join_keys: List[str],
        feature_view: FeatureView,
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def normalize_timestamp(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        timestamp_field: str,
        created_timestamp_column: str,
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def filter_ttl(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        feature_view: FeatureView,
        entity_df_event_timestamp_col: str,
        timestamp_field: str,
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def filter_time_range(
        self,
        source_df: Union[pd.DataFrame, Any],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def drop_duplicates(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        subset: List[str],
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def drop(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        columns_list: List[str],
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def add_static_column(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        column_name: str,
        column_value: str,
    ) -> Union[pd.DataFrame, Any]:
        ...

    @abstractmethod
    def select(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        columns_list: List[str]
    ) -> Union[pd.DataFrame, Any]:
        ...

    def merge(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        df_to_join: Union[pd.DataFrame, Any],
        join_keys: List[str],
        feature_view: FeatureView,
    ) -> Union[pd.DataFrame, Any]:
        # tmp join keys needed for cross join with null join table view
        tmp_join_keys = []
        if not join_keys:
            entity_df_with_features=self.add_static_column(entity_df_with_features, "__tmp", 1)
            df_to_join=self.add_static_column(df_to_join, "__tmp", 1)
            tmp_join_keys = ["__tmp"]

        # Get only data with requested entities
        if YUMMY_ALL in entity_df_with_features.columns:
            df_to_join=self.add_static_column(df_to_join, DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL, self.first_event_timestamp(entity_df_with_features, DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL))
        else:
            df_to_join = self.join(
                entity_df_with_features,
                df_to_join,
                join_keys or tmp_join_keys,
                feature_view,
            )

        if tmp_join_keys:
            df_to_join = self.drop(df_to_join, tmp_join_keys)

        return df_to_join

    def drop_df_duplicates(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        all_join_keys: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        entity_df_event_timestamp_col: Optional[str] = None,
    ) -> Union[pd.DataFrame, Any]:
        if created_timestamp_column:
            df_to_join =  self.sort_values(df_to_join, by=[created_timestamp_column, timestamp_field], na_position="first")
        else:
            df_to_join = self.sort_values(df_to_join, by=timestamp_field, na_position="first")

        if entity_df_event_timestamp_col:
            return self.drop_duplicates(df_to_join, subset=all_join_keys + [entity_df_event_timestamp_col])
        else:
            return self.drop_duplicates(df_to_join, subset=all_join_keys)

    def drop_columns(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        timestamp_field: str,
        created_timestamp_column: str,
    ) -> Union[pd.DataFrame, Any]:
        entity_df_with_features = self.drop(df_to_join, [timestamp_field])

        if created_timestamp_column:
            entity_df_with_features = self.drop(entity_df_with_features,[created_timestamp_column])

        return entity_df_with_features

    def field_mapping(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        feature_view: FeatureView,
        features: List[str],
        right_entity_key_columns: List[str],
        entity_df_event_timestamp_col: str,
        timestamp_field: str,
        full_feature_names: bool,
    ) -> Union[pd.DataFrame, Any]:
        # Rename columns by the field mapping dictionary if it exists
        if feature_view.batch_source.field_mapping:
            df_to_join = self.run_field_mapping(
                df_to_join, feature_view.batch_source.field_mapping
            )
        # Rename entity columns by the join_key_map dictionary if it exists
        if feature_view.projection.join_key_map:
            df_to_join = self.run_field_mapping(
                df_to_join, feature_view.projection.join_key_map
            )

        # Build a list of all the features we should select from this source
        feature_names = []
        columns_map = {}
        for feature in features:
            # Modify the separator for feature refs in column names to double underscore. We are using
            # double underscore as separator for consistency with other databases like BigQuery,
            # where there are very few characters available for use as separators
            if full_feature_names:
                formatted_feature_name = (
                    f"{feature_view.projection.name_to_use()}__{feature}"
                )
            else:
                formatted_feature_name = feature
            # Add the feature name to the list of columns
            feature_names.append(formatted_feature_name)
            columns_map[feature] = formatted_feature_name

        # Ensure that the source dataframe feature column includes the feature view name as a prefix
        df_to_join = self.run_field_mapping(df_to_join, columns_map)

        # Select only the columns we need to join from the feature dataframe
        df_to_join = self.select(df_to_join, right_entity_key_columns + feature_names)

        # Make sure to not have duplicated columns
        if entity_df_event_timestamp_col == timestamp_field:
            df_to_join = self.run_field_mapping(
                df_to_join, {timestamp_field: f"__{timestamp_field}"},
            )
            timestamp_field = f"__{timestamp_field}"

        return df_to_join, timestamp_field

    def get_entity_df_event_timestamp_range(
        self,
        entity_df: Union[pd.DataFrame, Any],
        entity_df_event_timestamp_col: str,
    ) -> Tuple[datetime, datetime]:
        """
        Finds min and max datetime in input entity_df data frame
        """
        if isinstance(entity_df, pd.DataFrame):
            entity_df_event_timestamp = entity_df.loc[entity_df_event_timestamp_col].infer_objects()
            if pd.api.types.is_string_dtype(entity_df_event_timestamp):
                entity_df_event_timestamp = pd.to_datetime(entity_df_event_timestamp, utc=True)

            return (
                entity_df_event_timestamp.min().to_pydatetime(),
                entity_df_event_timestamp.max().to_pydatetime(),
            )

        # TODO: check if this metadata is really needed
        return None, None

    def columns_list(
        self,
        entity_df: Union[pd.DataFrame, Any],
    ) -> List[str]:
        """
        Reads columns list
        """
        return entity_df.columns

    def select_dtypes_columns(
        self,
        entity_df: Union[pd.DataFrame, Any],
        include: List[str]
    ) -> List[str]:
        return entity_df.select_dtypes(include=include).columns

    def create_retrival_job(
        self,
        evaluation_function: Callable,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
    ) -> RetrievalJob:
        return self.retrival_job_type(
            evaluation_function=evaluation_function,
            full_feature_names=full_feature_names,
            on_demand_feature_views=on_demand_feature_views,
            metadata=metadata,
        )

    def read_datasource(
        self,
        data_source,
        features: List[str],
        entity_df: Optional[Union[pd.DataFrame, Any]] = None,
    ) -> Union[pyarrow.Table, pd.DataFrame, Any]:
        """
        Reads data source
        """
        assert issubclass(data_source.reader_type, YummyDataSourceReader)
        reader: YummyDataSourceReader = data_source.reader_type()
        return reader.read_datasource(data_source, features, self, entity_df)


class BackendFactory:

    @staticmethod
    def create(
        backend_type: BackendType,
        backend_config: BackendConfig) -> Backend:

        print(f'I will use {backend_type} backend')

        if backend_type == BackendType.dask:
            from yummy.backends.dask import DaskBackend
            return DaskBackend(backend_config)
        elif backend_type == BackendType.ray:
            import ray
            import dask
            from ray.util.dask import ray_dask_get
            ray.init(ignore_reinit_error=True)
            dask.config.set(scheduler=ray_dask_get)
            from yummy.backends.dask import DaskBackend
            return DaskBackend(backend_config)
        elif backend_type == BackendType.spark:
            from yummy.backends.spark import SparkBackend
            return SparkBackend(backend_config)
        elif backend_type == BackendType.polars:
            from yummy.backends.polars import PolarsBackend
            return PolarsBackend(backend_config)

        return PolarsBackend(backend_config)


class YummyDataSourceReader(ABC):

    @abstractmethod
    def read_datasource(
        self,
        data_source,
        features: List[str],
        backend: Backend,
        entity_df: Optional[Union[pd.DataFrame, Any]] = None,
    ) -> Union[pyarrow.Table, pd.DataFrame, Any]:
        ...


class YummyOfflineStoreConfig(BackendConfig):
    """Offline store config for local (file-based) store"""

    type: Literal["yummy.YummyOfflineStore"] = "yummy.YummyOfflineStore"
    """ Offline store type selector"""


class YummyOfflineStore(OfflineStore):
    @staticmethod
    @log_exceptions_and_usage(offline_store="yummy")
    def get_historical_features(
        config: RepoConfig,
        feature_views: List[FeatureView],
        feature_refs: List[str],
        entity_df: Union[pd.DataFrame, str],
        registry: Registry,
        project: str,
        full_feature_names: bool = False,
    ) -> RetrievalJob:

        backend_type = config.offline_store.backend
        backend = BackendFactory.create(backend_type, config.offline_store)
        entity_df = backend.prepare_entity_df(entity_df)
        all_columns = backend.columns_list(entity_df)

        entity_df_event_timestamp_col = DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL  # local modifiable copy of global variable
        if entity_df_event_timestamp_col not in all_columns:
            datetime_columns = backend.select_dtypes_columns(
                entity_df,
                include=["datetime", "datetimetz"]
            )
            if len(datetime_columns) == 1:
                print(
                    f"Using {datetime_columns[0]} as the event timestamp. To specify a column explicitly, please name it {DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL}."
                )
                entity_df_event_timestamp_col = datetime_columns[0]
            else:
                raise ValueError(
                    f"Please provide an entity_df with a column named {DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL} representing the time of events."
                )

        (
            feature_views_to_features,
            on_demand_feature_views_to_features,
        ) = _get_requested_feature_views_to_features_dict(
            feature_refs,
            feature_views,
            registry.list_on_demand_feature_views(config.project),
        )

        entity_df_event_timestamp_range = backend.get_entity_df_event_timestamp_range(
            entity_df, entity_df_event_timestamp_col
        )

        # Create lazy function that is only called from the RetrievalJob object
        def evaluate_historical_retrieval():

            entity_df_with_features = backend.normalize_timezone(entity_df, entity_df_event_timestamp_col)

            entity_df_with_features = backend.sort_values(entity_df_with_features, entity_df_event_timestamp_col)

            join_keys = []
            all_join_keys = []

            # Load feature view data from sources and join them incrementally
            for feature_view, features in feature_views_to_features.items():
                timestamp_field = (
                    feature_view.batch_source.timestamp_field
                )
                created_timestamp_column = (
                    feature_view.batch_source.created_timestamp_column
                )

                # Build a list of entity columns to join on (from the right table)
                join_keys = []

                for entity_name in feature_view.entities:
                    entity = registry.get_entity(entity_name, project)
                    join_key = feature_view.projection.join_key_map.get(
                        entity.join_key, entity.join_key
                    )
                    join_keys.append(join_key)

                right_entity_key_columns = [
                    timestamp_field,
                    created_timestamp_column,
                ] + join_keys
                right_entity_key_columns = [c for c in right_entity_key_columns if c]

                all_join_keys = list(set(all_join_keys + join_keys))

                df_to_join = backend.read_datasource(feature_view.batch_source, features, entity_df_with_features)

                df_to_join, timestamp_field = backend.field_mapping(
                    df_to_join,
                    feature_view,
                    features,
                    right_entity_key_columns,
                    entity_df_event_timestamp_col,
                    timestamp_field,
                    full_feature_names,
                )

                df_to_join = backend.merge(entity_df_with_features, df_to_join, join_keys, feature_view)

                df_to_join = backend.normalize_timestamp(
                    df_to_join, timestamp_field, created_timestamp_column
                )


                df_to_join = backend.filter_ttl(
                    df_to_join,
                    feature_view,
                    entity_df_event_timestamp_col,
                    timestamp_field,
                )

                df_to_join = backend.drop_df_duplicates(
                    df_to_join,
                    all_join_keys,
                    timestamp_field,
                    created_timestamp_column,
                    entity_df_event_timestamp_col,
                )

                entity_df_with_features = backend.drop_columns(
                    df_to_join, timestamp_field, created_timestamp_column
                )

                # Ensure that we delete dataframes to free up memory
                del df_to_join

            return entity_df_with_features

        job = backend.create_retrival_job(
            evaluation_function=evaluate_historical_retrieval,
            full_feature_names=full_feature_names,
            on_demand_feature_views=OnDemandFeatureView.get_requested_odfvs(
                feature_refs, project, registry
            ),
            metadata=RetrievalMetadata(
                features=feature_refs,
                keys=list(set(all_columns) - {entity_df_event_timestamp_col}),
                min_event_timestamp=entity_df_event_timestamp_range[0],
                max_event_timestamp=entity_df_event_timestamp_range[1],
            ),
        )
        return job

    @staticmethod
    @log_exceptions_and_usage(offline_store="yummy")
    def pull_latest_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:

        backend_type = config.offline_store.backend
        backend = BackendFactory.create(backend_type, config.offline_store)

        # Create lazy function that is only called from the RetrievalJob object
        def evaluate_offline_job():
            source_df = backend.read_datasource(data_source, feature_name_columns)

            source_df = backend.normalize_timestamp(
                source_df, timestamp_field, created_timestamp_column
            )

            all_columns = backend.columns_list(source_df)

            source_columns = set(all_columns)
            if not set(join_key_columns).issubset(source_columns):
                raise FeastJoinKeysDuringMaterialization(
                    data_source.path, set(join_key_columns), source_columns
                )

            ts_columns = (
                [timestamp_field, created_timestamp_column]
                if created_timestamp_column
                else [timestamp_field]
            )

            source_df = backend.filter_time_range(source_df, timestamp_field, start_date, end_date)

            columns_to_extract = set(
                join_key_columns + feature_name_columns + ts_columns
            )
            if join_key_columns:
                source_df = backend.drop_df_duplicates(source_df, join_key_columns, timestamp_field, created_timestamp_column)
            else:
                source_df = backend.add_static_column(source_df, DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL)
                columns_to_extract.add(DUMMY_ENTITY_ID)


            return backend.select(source_df, list(columns_to_extract))

        # When materializing a single feature view, we don't need full feature names. On demand transforms aren't materialized
        return backend.create_retrival_job(
            evaluation_function=evaluate_offline_job, full_feature_names=False,
        )

    @staticmethod
    @log_exceptions_and_usage(offline_store="file")
    def pull_all_from_table_or_query(
        config: RepoConfig,
        data_source: DataSource,
        join_key_columns: List[str],
        feature_name_columns: List[str],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> RetrievalJob:
        return YummyOfflineStore.pull_latest_from_table_or_query(
            config=config,
            data_source=data_source,
            join_key_columns=join_key_columns
            + [timestamp_field],  # avoid deduplication
            feature_name_columns=feature_name_columns,
            timestamp_field=timestamp_field,
            created_timestamp_column=None,
            start_date=start_date,
            end_date=end_date,
        )





import os
from .backends.backend import YummyOfflineStore, YummyOfflineStoreConfig, select_all
from .sources.file import ParquetSource, CsvSource
from .sources.delta import DeltaSource
from .sources.iceberg import IcebergSource
from .providers.provider import YummyProvider
from .registries.registry import YummyRegistryStore

class DeprecationHelper(object):
    def __init__(self, new_target, old_name):
        self.new_target = new_target
        self.old_name = old_name

    def _warn(self):
        from warnings import warn
        warn(f"The class {self.old_name} will be replaced with {self.new_target.__name__}")

    def __call__(self, *args, **kwargs):
        self._warn()
        return self.new_target(*args, **kwargs)

    def __getattr__(self, attr):
        self._warn()
        return getattr(self.new_target, attr)

CsvDataSource=DeprecationHelper(CsvSource, "CsvDataSource")
ParquetDataSource=DeprecationHelper(ParquetSource, "ParquetDataSource")
DeltaDataSource=DeprecationHelper(DeltaSource, "DeltaDataSource")
IcebergDataSource=DeprecationHelper(IcebergSource, "IcebergDataSource")

os.environ["FEAST_USAGE"]="False"
__all__ = ["YummyProvider",
           "YummyRegistryStore",
           "YummyOfflineStore",
           "YummyOfflineStoreConfig",
           "ParquetSource",
           "CsvSource",
           "DeltaSource",
           "IcebergSource",
           "ParquetDataSource",
           "CsvDataSource",
           "DeltaDataSource",
           "IcebergDataSource",
           "select_all"]



import uuid
import os
from datetime import datetime
from pathlib import Path

from feast.registry_store import RegistryStore
from feast.protos.feast.core.Registry_pb2 import Registry as RegistryProto
from feast.repo_config import RegistryConfig
from feast.usage import log_exceptions_and_usage

class YummyRegistryStore(RegistryStore):
    def __init__(self, registry_config: RegistryConfig, repo_path: Path):
        self._registry: RegistryStore = self._factory(registry_config, repo_path)

    def _factory(self, registry_config: RegistryConfig, repo_path: Path):
        path = registry_config.path

        if path.startswith("s3://"):
            from feast.infra.aws import S3RegistryStore
            if hasattr(registry_config, "s3_endpoint_override"):
                os.environ["FEAST_S3_ENDPOINT_URL"]=registry_config.s3_endpoint_override
            return S3RegistryStore(registry_config, repo_path)
        elif path.startswith("gs://"):
            from feast.infra.gcp import GCSRegistryStore
            return GCSRegistryStore(registry_config, repo_path)
        else:
            from feast.infra.local import LocalRegistryStore
            return LocalRegistryStore(registry_config, repo_path)

    @log_exceptions_and_usage(registry="local")
    def get_registry_proto(self):
        return self._registry.get_registry_proto()

    @log_exceptions_and_usage(registry="local")
    def update_registry_proto(self, registry_proto: RegistryProto):
        self._registry.update_registry_proto(registry_proto)

    def teardown(self):
        self._registry.teardown()

    def _write_registry(self, registry_proto: RegistryProto):
        self._registry._write_registry(registry_proto)




from feast.infra.local import LocalProvider
from feast.repo_config import RepoConfig
from feast.infra.offline_stores.offline_utils import get_offline_store_from_config
from yummy import YummyOfflineStoreConfig

class YummyProvider(LocalProvider):

    def __init__(self, config: RepoConfig):
        super().__init__(config)

        if hasattr(config, "backend"):
            config.offline_store.backend=config.backend

        if hasattr(config, "backend_config"):
            config.offline_store.config=config.backend_config






import json
from datetime import datetime
from typing import Callable, Dict, List, Optional, Union, Any

from feast import type_map
from feast.data_source import DataSource
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.repo_config import RepoConfig
from feast.value_type import ValueType
import pandas as pd
import pyarrow
from yummy.sources.source import YummyDataSource
from yummy.backends.backend import Backend, BackendType, YummyDataSourceReader


class YummyFileDataSource(YummyDataSource):
    """Custom data source class for local files"""

    def __init__(
        self,
        *,
        event_timestamp_column: Optional[str] = None,
        created_timestamp_column: Optional[str] = None,
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
        path: Optional[str] = None,
        s3_endpoint_override: Optional[str] = None,
    ):
        super().__init__(
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
            name=name,
            timestamp_field=timestamp_field,
        )
        self._path = path
        self._s3_endpoint_override = s3_endpoint_override

    @property
    def reader_type(self):
        """
        Returns the reader type which will read data source
        """
        ...

    @property
    def path(self):
        """
        Returns the file path of this feature data source.
        """
        return self._path

    @property
    def s3_endpoint_override(self):
        """
        Returns the s3_endpoint_override of this feature data source.
        """
        return self._s3_endpoint_override

    def to_proto(self) -> DataSourceProto:
        """
        Creates a DataSource proto representation of this object, by serializing some
        custom options into the custom_options field as a binary encoded json string.
        """
        config_json = json.dumps(
            {"path": self.path, "s3_endpoint_override": self.s3_endpoint_override}
        )
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.CUSTOM_SOURCE,
            field_mapping=self.field_mapping,
            description=self.description,
            tags=self.tags,
            owner=self.owner,
            timestamp_field=self.timestamp_field,
            created_timestamp_column=self.created_timestamp_column,
            custom_options=DataSourceProto.CustomSourceOptions(
                configuration=bytes(config_json, encoding="utf8")
            ),
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    def get_table_query_string(self) -> str:
        pass

    def validate(self, config: RepoConfig):
        # TODO: validate a FileSource
        pass

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.pa_to_feast_value_type


class ParquetSource(YummyFileDataSource):
    """Custom data source class for parquets files"""

    def __init__(
        self,
        *,
        event_timestamp_column: Optional[str] = None,
        created_timestamp_column: Optional[str] = None,
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
        path: Optional[str] = None,
        s3_endpoint_override: Optional[str] = None,
    ):
        super().__init__(
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
            name=name,
            timestamp_field=timestamp_field,
            path=path,
            s3_endpoint_override=s3_endpoint_override,
        )

    @property
    def reader_type(self):
        """
        Returns the reader type which will read data source
        """
        return ParquetSourceReader

    @staticmethod
    def from_proto(data_source: DataSourceProto) -> Any:
        """
        Creates a `CustomFileDataSource` object from a DataSource proto, by
        parsing the CustomSourceOptions which is encoded as a binary json string.
        """
        custom_source_options = str(
            data_source.custom_options.configuration, encoding="utf8"
        )
        path = json.loads(custom_source_options)["path"]
        s3_endpoint_override = json.loads(custom_source_options)["s3_endpoint_override"]
        return ParquetSource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
            path=path,
            s3_endpoint_override=s3_endpoint_override,
        )


class CsvSource(YummyFileDataSource):
    """Custom data source class for parquets files"""

    def __init__(
        self,
        *,
        event_timestamp_column: Optional[str] = None,
        created_timestamp_column: Optional[str] = None,
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
        path: Optional[str] = None,
        s3_endpoint_override: Optional[str] = None,
        delimiter: Optional[str] = ',',
        header: Optional[bool] = True,
    ):
        super().__init__(
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
            name=name,
            timestamp_field=timestamp_field,
            path=path,
            s3_endpoint_override=s3_endpoint_override,
        )
        self._delimiter = delimiter
        self._header = header

    @property
    def reader_type(self):
        """
        Returns the reader type which will read data source
        """
        return CsvSourceReader

    @property
    def delimiter(self):
        return self._delimiter

    @property
    def header(self):
        return self._header

    def to_proto(self) -> DataSourceProto:
        """
        Creates a DataSource proto representation of this object, by serializing some
        custom options into the custom_options field as a binary encoded json string.
        """
        config_json = json.dumps(
            {"path": self.path,
             "s3_endpoint_override": self.s3_endpoint_override,
             "delimiter": self.delimiter,
             "header": self.header,
             }
        )
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.CUSTOM_SOURCE,
            field_mapping=self.field_mapping,
            description=self.description,
            tags=self.tags,
            owner=self.owner,
            timestamp_field=self.timestamp_field,
            created_timestamp_column=self.created_timestamp_column,
            custom_options=DataSourceProto.CustomSourceOptions(
                configuration=bytes(config_json, encoding="utf8")
            ),
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    @staticmethod
    def from_proto(data_source: DataSourceProto) -> Any:
        """
        Creates a `CustomFileDataSource` object from a DataSource proto, by
        parsing the CustomSourceOptions which is encoded as a binary json string.
        """
        custom_source_options = str(
            data_source.custom_options.configuration, encoding="utf8"
        )
        path = json.loads(custom_source_options)["path"]
        s3_endpoint_override = json.loads(custom_source_options)["s3_endpoint_override"]
        delimiter = json.loads(custom_source_options)["delimiter"]
        header = json.loads(custom_source_options)["header"]
        return CsvSource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
            path=path,
            s3_endpoint_override=s3_endpoint_override,
            delimiter=delimiter,
            header=header,
        )


class ParquetSourceReader(YummyDataSourceReader):

    def read_datasource(
        self,
        data_source,
        features: List[str],
        backend: Backend,
        entity_df: Optional[Union[pd.DataFrame, Any]] = None,
    ) -> Union[pyarrow.Table, pd.DataFrame, Any]:
        backend_type = backend.backend_type
        if backend_type == BackendType.spark:
            from yummy.backends.spark import SparkBackend
            spark_backend: SparkBackend = backend
            spark_session = spark_backend.spark_session
            #TODO: add s3 endpoint override support
            path = data_source.path
            if "s3://" in path:
                path = path.replace('s3://','s3a://')

            return spark_session.read.parquet(path)

        elif backend_type in [BackendType.ray, BackendType.dask]:
            import dask.dataframe as dd
        elif backend_type == BackendType.polars:
            import polars as dd

        return dd.read_parquet(data_source.path, storage_options=self._storage_options(data_source),)

    def _storage_options(self, data_source):
        return (
            {
                "client_kwargs": {
                    "endpoint_url": data_source.s3_endpoint_override
                }
            }
            if data_source.s3_endpoint_override
            else None
        )


class CsvSourceReader(ParquetSourceReader):

    def read_datasource(
        self,
        data_source,
        features: List[str],
        backend: Backend,
        entity_df: Optional[Union[pd.DataFrame, Any]] = None,
    ) -> Union[pyarrow.Table, pd.DataFrame, Any]:
        backend_type = backend.backend_type
        timestamp_field = data_source.timestamp_field
        created_timestamp_column = data_source.created_timestamp_column

        if backend_type == BackendType.spark:
            from yummy.backends.spark import SparkBackend
            spark_backend: SparkBackend = backend
            spark_session = spark_backend.spark_session
            #TODO: add s3 endpoint override support
            path = data_source.path
            if "s3://" in path:
                path = path.replace('s3://','s3a://')

            return spark_session.read.options(header=data_source.header, delimeter=data_source.delimiter).csv(path)

        elif backend_type in [BackendType.ray, BackendType.dask]:
            import dask.dataframe as dd
            csv_df = dd.read_csv(
                data_source.path,
                storage_options=self._storage_options(data_source),
                delimiter=data_source.delimiter,
            )
            csv_df[timestamp_field] = dd.to_datetime(csv_df[timestamp_field])
            if created_timestamp_column:
                csv_df[created_timestamp_column] = dd.to_datetime(csv_df[created_timestamp_column])

            return csv_df
        elif backend_type == BackendType.polars:
            import polars as dd
            #TODO: check
            # csv_df = csv_df.with_column(pl.col("datetime").str.strptime(pl.Datetime, "%Y-%m-%d %H:%M:%S.%f", strict=False))
            return dd.read_csv(
                data_source.path,
                storage_options=self._storage_options(data_source),
                parse_dates=True,
                has_header=data_source.header,
                delimiter=data_source.delimiter,
            )







from datetime import datetime
from typing import Callable, List, Optional, Tuple, Union, Dict, Any

import pandas as pd
import pyarrow
import pyspark.sql.dataframe as sd
import pytz
from feast import FileSource, OnDemandFeatureView
from feast.data_source import DataSource
from feast.errors import FeastJoinKeysDuringMaterialization
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.offline_stores.offline_utils import (
    DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL,
)
from feast.infra.provider import _get_requested_feature_views_to_features_dict
from feast.registry import Registry
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.usage import log_exceptions_and_usage
from pydantic.typing import Literal
from pyspark.sql.functions import col, expr, lit
from pyspark.sql import SparkSession
from pyspark import SparkConf
from yummy.backends.backend import Backend, BackendType, BackendConfig

class SparkBackend(Backend):
    def __init__(self, backend_config: BackendConfig):
        super().__init__(backend_config)
        self._spark_session = self._get_spark_session(backend_config)

    @property
    def backend_type(self) -> BackendType:
        return BackendType.spark

    @property
    def retrival_job_type(self):
        return SparkRetrievalJob

    @property
    def spark_session(self) -> SparkSession:
        return self._spark_session

    def first_event_timestamp(
        self,
        entity_df: Union[pd.DataFrame, Any],
        column_name: str
    ) -> datetime:
        """
        Fetch first event timestamp
        """
        return entity_df.toPandas()[column_name][0]

    def prepare_entity_df(
        self,
        entity_df: Union[pd.DataFrame, Any],
    ) -> Union[pd.DataFrame, Any]:
        """
        Maps entity_df to type required by backend and finds event timestamp column
        """
        if not isinstance(entity_df, pd.DataFrame) and not isinstance(
            entity_df, sd.DataFrame
        ):
            raise ValueError(
                f"Please provide an entity_df of type {type(pd.DataFrame)} instead of type {type(entity_df)}"
            )

        if isinstance(entity_df, pd.DataFrame):
            entity_df: sd.DataFrame = self.spark_session.createDataFrame(entity_df)

        return entity_df

    def normalize_timezone(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        entity_df_event_timestamp_col: str,
    ) -> Union[pd.DataFrame, Any]:
        """
        Normalize timezon of input entity df to UTC
        """
        return entity_df_with_features

    def sort_values(
        self,
        entity_df: Union[pd.DataFrame, Any],
        by: Union[str,List[str]],
        ascending: bool = True,
        na_position: Optional[str] = "last",
    ) -> Union[pd.DataFrame, Any]:
        """
        Sorts entity df by selected column
        """

        if type(by) is str:
            by = [by]

        cols = [col(b) for b in by]

        if ascending and na_position=='first':
            cols = [col(b).asc_nulls_first() for b in by]
        elif ascending and na_position=='last':
            cols = [col(b).asc_nulls_last() for b in by]
        elif not ascending and na_position=='first':
            cols = [col(b).desc_nulls_first() for b in by]
        elif not ascending and na_position=='last':
            cols = [col(b).desc_nulls_last() for b in by]

        return entity_df.orderBy(cols)

    def run_field_mapping(
        self,
        table: Union[pd.DataFrame,Any],
        field_mapping: Dict[str, str],
    ):
        if field_mapping:
            return table.select(
                [col(c).alias(field_mapping.get(c, c)) for c in table.columns]
            )
        else:
            return table

    def join(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        df_to_join: Union[pd.DataFrame, Any],
        join_keys: List[str],
        feature_view: FeatureView,
    ) -> Union[pd.DataFrame, Any]:
        range_join = None
        if hasattr(feature_view.batch_source, 'range_join'):
            range_join = feature_view.batch_source.range_join

        if range_join:
            df_to_join = df_to_join.hint("range_join", range_join)

        df_to_join_cols = df_to_join.columns
        entity_df_with_features_columns = [c for c in entity_df_with_features.columns if c not in df_to_join_cols]
        cols = df_to_join_cols+entity_df_with_features_columns

        return entity_df_with_features.join(
            other=df_to_join,
            on=join_keys,
            how="left"
        ).select(*cols)

    def normalize_timestamp(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        timestamp_field: str,
        created_timestamp_column: str,
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join

    def filter_ttl(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        feature_view: FeatureView,
        entity_df_event_timestamp_col: str,
        timestamp_field: str,
    ) -> Union[pd.DataFrame, Any]:
        # Filter rows by defined timestamp tolerance
        if feature_view.ttl and feature_view.ttl.total_seconds() != 0:
            ttl_seconds = feature_view.ttl.total_seconds()
            df_to_join= df_to_join.filter(
                (
                    col(timestamp_field)
                    >= col(entity_df_event_timestamp_col)
                    - expr(f"INTERVAL {ttl_seconds} seconds")
                )
                & (
                    col(timestamp_field)
                    <= col(entity_df_event_timestamp_col)
                )
            )

        return df_to_join

    def filter_time_range(
        self,
        source_df: Union[pd.DataFrame, Any],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> Union[pd.DataFrame, Any]:
        return source_df.filter(
            (col(timestamp_field) >= start_date)
            & (col(timestamp_field) < end_date)
        )

    def drop_duplicates(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        subset: List[str],
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join.dropDuplicates(subset)

    def drop(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        columns_list: List[str],
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join.drop(*columns_list)

    def add_static_column(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        column_name: str,
        column_value: str,
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join.withColumn(column_name, lit(column_value))

    def select(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        columns_list: List[str]
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join.select(*columns_list)

    def drop_df_duplicates(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        all_join_keys: List[str],
        timestamp_field: str,
        created_timestamp_column: Optional[str],
        entity_df_event_timestamp_col: Optional[str] = None,
    ) -> Union[pd.DataFrame, Any]:
        # This must be overriten and df must be reversed because in pyspark
        # there is no keep last in dropDuplicates
        if created_timestamp_column:
            df_to_join =  self.sort_values(df_to_join, ascending=False, by=[created_timestamp_column, timestamp_field], na_position="last")
        else:
            df_to_join = self.sort_values(df_to_join, ascending=False, by=timestamp_field, na_position="last")

        if entity_df_event_timestamp_col:
            return self.drop_duplicates(df_to_join,subset=all_join_keys + [entity_df_event_timestamp_col])
        else:
            return self.drop_duplicates(df_to_join,subset=all_join_keys)

    def _get_spark_session(
        self,
        backend_config: BackendConfig,
    ) -> SparkSession:
        spark_session = SparkSession.getActiveSession()

        if not spark_session:
            spark_builder = SparkSession.builder
            spark_conf = backend_config.config

            if spark_conf:
                spark_builder = spark_builder.config(
                    conf=SparkConf().setAll(spark_conf.items())
                )  # noqa

            spark_session = spark_builder.getOrCreate()

        return spark_session


class SparkRetrievalJob(RetrievalJob):
    def __init__(
        self,
        evaluation_function: Callable,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
    ):
        """Initialize a lazy historical retrieval job"""

        # The evaluation function executes a stored procedure to compute a historical retrieval.
        self.evaluation_function = evaluation_function
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = (
            on_demand_feature_views if on_demand_feature_views else []
        )
        self._metadata = metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> Optional[List[OnDemandFeatureView]]:
        return self._on_demand_feature_views

    @log_exceptions_and_usage
    def _to_df_internal(self) -> pd.DataFrame:
        # Only execute the evaluation function to build the final historical retrieval dataframe at the last moment.
        df = self.evaluation_function().toPandas()
        return df

    @log_exceptions_and_usage
    def _to_arrow_internal(self):
        # Only execute the evaluation function to build the final historical retrieval dataframe at the last moment.
        df = self.evaluation_function().toPandas()
        return pyarrow.Table.from_pandas(df)

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        self._metadata

    def persist(self, storage: SavedDatasetStorage):
        pass





from datetime import datetime
from typing import Callable, List, Optional, Tuple, Union, Dict, Any

import dask.dataframe as dd
import pandas as pd
import pyarrow
import pytz
from pydantic.typing import Literal

from feast import FileSource, OnDemandFeatureView
from feast.data_source import DataSource
from feast.errors import FeastJoinKeysDuringMaterialization
from feast.feature_view import DUMMY_ENTITY_ID, DUMMY_ENTITY_VAL, FeatureView
from feast.infra.offline_stores.file_source import SavedDatasetFileStorage
from feast.infra.offline_stores.offline_store import (
    OfflineStore,
    RetrievalJob,
    RetrievalMetadata,
)
from feast.infra.offline_stores.offline_utils import (
    DEFAULT_ENTITY_DF_EVENT_TIMESTAMP_COL,
)
from feast.infra.provider import (
    _get_requested_feature_views_to_features_dict,
)
from feast.registry import Registry
from feast.repo_config import FeastConfigBaseModel, RepoConfig
from feast.saved_dataset import SavedDatasetStorage
from feast.usage import log_exceptions_and_usage
from yummy.backends.backend import Backend, BackendType, BackendConfig


class DaskBackend(Backend):
    def __init__(self, backend_config: BackendConfig):
        super().__init__(backend_config)

    @property
    def backend_type(self) -> BackendType:
        return BackendType.dask

    @property
    def retrival_job_type(self):
        return DaskRetrievalJob

    def prepare_entity_df(
        self,
        entity_df: Union[pd.DataFrame, Any],
    ) -> Union[pd.DataFrame, Any]:
        """
        Maps entity_df to type required by backend and finds event timestamp column
        """
        if not isinstance(entity_df, pd.DataFrame) and not isinstance(
            entity_df, dd.DataFrame
        ):
            raise ValueError(
                f"Please provide an entity_df of type {type(pd.DataFrame)} instead of type {type(entity_df)}"
            )

        if isinstance(entity_df, pd.DataFrame):
            entity_df: dd.DataFrame = dd.from_pandas(entity_df, npartitions=1)

        return entity_df

    def first_event_timestamp(
        self,
        entity_df: Union[pd.DataFrame, Any],
        column_name: str
    ) -> datetime:
        """
        Fetch first event timestamp
        """
        return entity_df.compute()[column_name][0]


    def normalize_timezone(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        entity_df_event_timestamp_col: str,
    ) -> Union[pd.DataFrame, Any]:
        """
        Normalize timezon of input entity df to UTC
        """
        entity_df_event_timestamp_col_type = entity_df_with_features.dtypes[
            entity_df_event_timestamp_col
        ]
        if (
            not hasattr(entity_df_event_timestamp_col_type, "tz")
            or entity_df_event_timestamp_col_type.tz != pytz.UTC
        ):
            # Make sure all event timestamp fields are tz-aware. We default tz-naive fields to UTC
            entity_df_with_features[
                entity_df_event_timestamp_col
            ] = entity_df_with_features[entity_df_event_timestamp_col].apply(
                lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc),
                meta=(entity_df_event_timestamp_col, "datetime64[ns, UTC]"),
            )

            # Convert event timestamp column to datetime and normalize time zone to UTC
            # This is necessary to avoid issues with pd.merge_asof
            if isinstance(entity_df_with_features, dd.DataFrame):
                entity_df_with_features[
                    entity_df_event_timestamp_col
                ] = dd.to_datetime(
                    entity_df_with_features[entity_df_event_timestamp_col], utc=True
                )
            else:
                entity_df_with_features[
                    entity_df_event_timestamp_col
                ] = pd.to_datetime(
                    entity_df_with_features[entity_df_event_timestamp_col], utc=True
                )

        return entity_df_with_features

    def sort_values(
        self,
        entity_df: Union[pd.DataFrame, Any],
        by: Union[str,List[str]],
        ascending: bool = True,
        na_position: Optional[str] = "last",
    ) -> Union[pd.DataFrame, Any]:
        """
        Sorts entity df by selected column
        """
        return entity_df.sort_values(by, ascending=ascending, na_position=na_position).persist()

    def run_field_mapping(
        self,
        table: Union[pd.DataFrame,Any],
        field_mapping: Dict[str, str],
    ):
        if field_mapping:
            # run field mapping in the forward direction
            table = table.rename(columns=field_mapping)
            table = table.persist()

        return table

    def join(
        self,
        entity_df_with_features: Union[pd.DataFrame, Any],
        df_to_join: Union[pd.DataFrame, Any],
        join_keys: List[str],
        feature_view: FeatureView,
    ) -> Union[pd.DataFrame, Any]:
        return dd.merge(
                entity_df_with_features,
                df_to_join,
                left_on=join_keys,
                right_on=join_keys,
                suffixes=("", "__"),
                how="left",
            )

    def normalize_timestamp(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        timestamp_field: str,
        created_timestamp_column: str,
    ) -> Union[pd.DataFrame, Any]:
        df_to_join_types = df_to_join.dtypes
        timestamp_field_type = df_to_join_types[timestamp_field]

        if created_timestamp_column:
            created_timestamp_column_type = df_to_join_types[created_timestamp_column]

        if (
            not hasattr(timestamp_field_type, "tz")
            or timestamp_field_type.tz != pytz.UTC
        ):
            # Make sure all timestamp fields are tz-aware. We default tz-naive fields to UTC
            df_to_join[timestamp_field] = df_to_join[timestamp_field].apply(
                lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc),
                meta=(timestamp_field, "datetime64[ns, UTC]"),
            )

        if created_timestamp_column and (
            not hasattr(created_timestamp_column_type, "tz")
            or created_timestamp_column_type.tz != pytz.UTC
        ):
            df_to_join[created_timestamp_column] = df_to_join[
                created_timestamp_column
            ].apply(
                lambda x: x if x.tzinfo is not None else x.replace(tzinfo=pytz.utc),
                meta=(timestamp_field, "datetime64[ns, UTC]"),
            )

        return df_to_join.persist()

    def filter_ttl(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        feature_view: FeatureView,
        entity_df_event_timestamp_col: str,
        timestamp_field: str,
    ) -> Union[pd.DataFrame, Any]:
        # Filter rows by defined timestamp tolerance
        if feature_view.ttl and feature_view.ttl.total_seconds() != 0:
            df_to_join = df_to_join[
                (
                    df_to_join[timestamp_field]
                    >= df_to_join[entity_df_event_timestamp_col] - feature_view.ttl
                )
                & (
                    df_to_join[timestamp_field]
                    <= df_to_join[entity_df_event_timestamp_col]
                )
            ]

            df_to_join = df_to_join.persist()

        return df_to_join

    def filter_time_range(
        self,
        source_df: Union[pd.DataFrame, Any],
        timestamp_field: str,
        start_date: datetime,
        end_date: datetime,
    ) -> Union[pd.DataFrame, Any]:
        source_df = source_df[
            (source_df[timestamp_field] >= start_date)
            & (source_df[timestamp_field] < end_date)
        ]

        return source_df.persist()

    def drop_duplicates(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        subset: List[str],
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join.drop_duplicates(subset, keep='last', ignore_index=True,).persist()

    def drop(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        columns_list: List[str],
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join.drop(columns_list, axis=1).persist()

    def add_static_column(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        column_name: str,
        column_value: str,
    ) -> Union[pd.DataFrame, Any]:
        df_to_join[column_name]=column_value
        return df_to_join

    def select(
        self,
        df_to_join: Union[pd.DataFrame, Any],
        columns_list: List[str]
    ) -> Union[pd.DataFrame, Any]:
        return df_to_join[columns_list].persist()

class DaskRetrievalJob(RetrievalJob):
    def __init__(
        self,
        evaluation_function: Callable,
        full_feature_names: bool,
        on_demand_feature_views: Optional[List[OnDemandFeatureView]] = None,
        metadata: Optional[RetrievalMetadata] = None,
    ):
        """Initialize a lazy historical retrieval job"""

        # The evaluation function executes a stored procedure to compute a historical retrieval.
        self.evaluation_function = evaluation_function
        self._full_feature_names = full_feature_names
        self._on_demand_feature_views = (
            on_demand_feature_views if on_demand_feature_views else []
        )
        self._metadata = metadata

    @property
    def full_feature_names(self) -> bool:
        return self._full_feature_names

    @property
    def on_demand_feature_views(self) -> Optional[List[OnDemandFeatureView]]:
        return self._on_demand_feature_views

    @log_exceptions_and_usage
    def _to_df_internal(self) -> pd.DataFrame:
        # Only execute the evaluation function to build the final historical retrieval dataframe at the last moment.
        df = self.evaluation_function().compute()
        return df

    @log_exceptions_and_usage
    def _to_arrow_internal(self):
        # Only execute the evaluation function to build the final historical retrieval dataframe at the last moment.
        df = self.evaluation_function().compute()
        return pyarrow.Table.from_pandas(df)

    def persist(self, storage: SavedDatasetStorage):
        assert isinstance(storage, SavedDatasetFileStorage)

        filesystem, path = FileSource.create_filesystem_and_path(
            storage.file_options.file_url, storage.file_options.s3_endpoint_override,
        )

        if path.endswith(".parquet"):
            pyarrow.parquet.write_table(
                self.to_arrow(), where=path, filesystem=filesystem
            )
        else:
            # otherwise assume destination is directory
            pyarrow.parquet.write_to_dataset(
                self.to_arrow(), root_path=path, filesystem=filesystem
            )

    @property
    def metadata(self) -> Optional[RetrievalMetadata]:
        return self._metadata





import json
from datetime import datetime
from typing import Callable, Dict, List, Optional, Union, Any
import pandas as pd
import pyarrow
from feast import type_map
from feast.data_source import DataSource
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.repo_config import RepoConfig
from feast.value_type import ValueType
from yummy.sources.source import YummyDataSource
from yummy.backends.backend import Backend, BackendType, YummyDataSourceReader

class DeltaSource(YummyDataSource):
    """Custom data source class for local files"""

    def __init__(
        self,
        *,
        event_timestamp_column: Optional[str] = None,
        created_timestamp_column: Optional[str] = None,
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
        path: Optional[str] = None,
        s3_endpoint_override: Optional[str] = None,
        range_join: Optional[int] = None,
    ):
        super().__init__(
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
            name=name,
            timestamp_field=timestamp_field,
        )
        self._path = path
        self._s3_endpoint_override = s3_endpoint_override
        self._range_join = range_join

    @property
    def reader_type(self):
        """
        Returns the reader type which will read data source
        """
        return DeltaSourceReader

    @property
    def path(self):
        """
        Returns the file path of this feature data source.
        """
        return self._path

    @property
    def range_join(self):
        """
        Returns the file path of this feature data source.
        """
        return self._range_join

    @property
    def s3_endpoint_override(self):
        """
        Returns the s3_endpoint_override of this feature data source.
        """
        return self._s3_endpoint_override

    @staticmethod
    def from_proto(data_source: DataSourceProto):
        """
        Creates a `CustomFileDataSource` object from a DataSource proto, by
        parsing the CustomSourceOptions which is encoded as a binary json string.
        """
        custom_source_options = str(
            data_source.custom_options.configuration, encoding="utf8"
        )
        path = json.loads(custom_source_options)["path"]
        s3_endpoint_override = json.loads(custom_source_options)["s3_endpoint_override"]
        range_join = json.loads(custom_source_options)["range_join"]
        return DeltaSource(
            name=data_source.name,
            field_mapping=dict(data_source.field_mapping),
            timestamp_field=data_source.timestamp_field,
            created_timestamp_column=data_source.created_timestamp_column,
            description=data_source.description,
            tags=dict(data_source.tags),
            owner=data_source.owner,
            path=path,
            s3_endpoint_override=s3_endpoint_override,
            range_join=range_join,
        )

    def to_proto(self) -> DataSourceProto:
        """
        Creates a DataSource proto representation of this object, by serializing some
        custom options into the custom_options field as a binary encoded json string.
        """
        config_json = json.dumps(
            {"path": self.path, "s3_endpoint_override": self.s3_endpoint_override, "range_join": self.range_join}
        )
        data_source_proto = DataSourceProto(
            name=self.name,
            type=DataSourceProto.CUSTOM_SOURCE,
            field_mapping=self.field_mapping,
            description=self.description,
            tags=self.tags,
            owner=self.owner,
            timestamp_field=self.timestamp_field,
            created_timestamp_column=self.created_timestamp_column,
            custom_options=DataSourceProto.CustomSourceOptions(
                configuration=bytes(config_json, encoding="utf8")
            ),
        )

        data_source_proto.timestamp_field = self.timestamp_field
        data_source_proto.created_timestamp_column = self.created_timestamp_column

        return data_source_proto

    def get_table_query_string(self) -> str:
        pass

    def validate(self, config: RepoConfig):
        # TODO: validate a DeltaSource
        pass

    @staticmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        return type_map.pa_to_feast_value_type


class DeltaSourceReader(YummyDataSourceReader):

    def read_datasource(
        self,
        data_source,
        features: List[str],
        backend: Backend,
        entity_df: Optional[Union[pd.DataFrame, Any]] = None,
    ) -> Union[pyarrow.Table, pd.DataFrame, Any]:
        backend_type = backend.backend_type
        path = data_source.path
        if backend_type == BackendType.spark:
            from yummy.backends.spark import SparkBackend
            spark_backend: SparkBackend = backend
            spark_session = spark_backend.spark_session
            #TODO: add s3 endpoint override support
            path = data_source.path
            if "s3://" in path:
                path = path.replace('s3://','s3a://')
            return spark_session.read.format('delta').load(path)
        elif backend_type in [BackendType.ray, BackendType.dask]:
            import dask.dataframe as dd
            from deltalake import DeltaTable
            dt = DeltaTable(path)
            return dd.from_pandas(dt.to_pandas(), npartitions=1)
        elif backend_type == BackendType.polars:
            import polars as pl
            from deltalake import DeltaTable
            dt = DeltaTable(path)
            return pl.DataFrame(dt.to_pyarrow_table())

        raise NotImplementedError(f'Delta lake support not implemented for {backend_type}')




import click
import yummy_rs
import yummy_mlflow


@click.group()
@click.pass_context
def cli(
    ctx: click.Context
):
    pass

@cli.command("serve")
@click.option("--host","-h",
              type=click.STRING,
              default="127.0.0.1",
              show_default=True,
              help='Host for the feature server.'
              )
@click.option("--port","-p",
              type=click.INT,
              default=6566,
              show_default=True,
              help='Port for the feature server.'
              )
@click.option("--filename","-f",
              type=click.STRING,
              default="feature_store.yaml",
              show_default=True,
              help='Configuration filename for the feature server.'
              )
@click.option("--log-level",
              type=click.STRING,
              default="error",
              show_default=True,
              help='Log level for the feature server. One of: debug, info, error, warn'
              )
@click.pass_context
def serve_command(
    ctx: click.Context,
    host: str,
    port: int,
    filename: str,
    log_level: str,
):
    """Start rust feature server."""
    yummy_rs.serve(filename, host, port, log_level)

@cli.group(name='models')
def models_cmd():
    """Models related commands"""
    pass

@models_cmd.command("serve")
@click.option("--host","-h",
              type=click.STRING,
              default="127.0.0.1",
              show_default=True,
              help='Host for the feature server.'
              )
@click.option("--port","-p",
              type=click.INT,
              default=6566,
              show_default=True,
              help='Port for the feature server.'
              )
@click.option("--model-uri","-m",
              type=click.STRING,
              required=True,
              help='Local model path.'
              )
@click.option("--log-level",
              type=click.STRING,
              default="error",
              show_default=True,
              help='Log level for the feature server. One of: debug, info, error, warn'
              )
@click.pass_context
def serve_command(
    ctx: click.Context,
    host: str,
    port: int,
    model_uri: str,
    log_level: str,
):
    """Start rust feature server."""
    yummy_mlflow.model_serve(model_uri, host, port, log_level)


if __name__ == '__main__':
    cli()



import json
from datetime import datetime
from typing import Callable, Dict, List, Optional, Union, Iterable, Tuple, Any
from abc import ABC, abstractmethod
from feast import type_map
from feast.data_source import DataSource
from feast.protos.feast.core.DataSource_pb2 import DataSource as DataSourceProto
from feast.repo_config import RepoConfig
from feast.value_type import ValueType
import pyarrow
import pandas as pd

class YummyDataSource(DataSource):
    """Custom data source class for local files"""

    def __init__(
        self,
        *,
        event_timestamp_column: Optional[str] = None,
        created_timestamp_column: Optional[str] = None,
        field_mapping: Optional[Dict[str, str]] = None,
        date_partition_column: Optional[str] = None,
        description: Optional[str] = "",
        tags: Optional[Dict[str, str]] = None,
        owner: Optional[str] = "",
        name: Optional[str] = None,
        timestamp_field: Optional[str] = None,
    ):
        super().__init__(
            event_timestamp_column=event_timestamp_column,
            created_timestamp_column=created_timestamp_column,
            field_mapping=field_mapping,
            date_partition_column=date_partition_column,
            description=description,
            tags=tags,
            owner=owner,
            name=name,
            timestamp_field=timestamp_field,
        )

    @property
    def reader_type(self):
        """
        Returns the reader type which will read data source
        """
        raise NotImplementedError("Reader type not defined")

    @staticmethod
    def from_proto(data_source: DataSourceProto) -> Any:
        """
        Creates a `CustomFileDataSource` object from a DataSource proto, by
        parsing the CustomSourceOptions which is encoded as a binary json string.
        """
        raise NotImplementedError

    def to_proto(self) -> DataSourceProto:
        """
        Creates a DataSource proto representation of this object, by serializing some
        custom options into the custom_options field as a binary encoded json string.
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def source_datatype_to_feast_value_type() -> Callable[[str], ValueType]:
        """
        Returns the callable method that returns Feast type given the raw column type.
        """
        raise NotImplementedError

    def get_table_column_names_and_types(
        self, config: RepoConfig
    ) -> Iterable[Tuple[str, str]]:
        """
        Returns the list of column names and raw column types.
        Args:
            config: Configuration object used to configure a feature store.
        """
        raise NotImplementedError

    def get_table_query_string(self) -> str:
        """
        Returns a string that can directly be used to reference this table in SQL.
        """
        raise NotImplementedError





from setuptools import find_packages, setup

NAME = "yummy"
REQUIRES_PYTHON = ">=3.7.0"

INSTALL_REQUIRE = [
    "feast~=0.22.1",
    "polars>=0.13.18",
    "yummy-rs~=0.0.7",
    "yummy-mlflow~=0.0.7",
]

DASK_REQUIRE = [
    "dask[distributed]>=2021.11.0",
]

RAY_REQUIRE = DASK_REQUIRE + [
    "ray[default,data]>=1.9.1",
]

SPARK_REQUIRE = [
    "pyspark>=3.0.0",
]

DELTA_REQUIRE = [
    "deltalake>=0.5.6",
]

DEV_REQUIRE = RAY_REQUIRE + \
    SPARK_REQUIRE + \
    DELTA_REQUIRE + [
    "flake8",
    "black==21.10b0",
    "isort>=5",
    "mypy>=0.790",
    "build>=0.7.0",
    "twine>=3.4.2",
    "pytest>=6.0.0",
]

setup(
    name=NAME,
    version="0.0.7",
    author="Qooba",
    long_description=open("README.md").read(),
    long_description_content_type="text/markdown",
    python_requires=REQUIRES_PYTHON,
    url="https://github.com/yummyml/yummy",
    packages=find_packages(exclude=["test*", "feature_repo*"]),
    install_requires=INSTALL_REQUIRE,
    extras_require={
        "dev": DEV_REQUIRE,
        "dask": DASK_REQUIRE,
        "ray": RAY_REQUIRE,
        "spark": SPARK_REQUIRE,
        "delta": DELTA_REQUIRE,
    },
    keywords=("feast featurestore polars dask ray pyspark offlinestore"),
    license='Apache License, Version 2.0',
    classifiers=[
        "License :: OSI Approved :: Apache Software License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
    ],
    entry_points={"console_scripts": ["yummy=yummy.cli:cli"]},
)
